@@ -12,16 +12,31 @@
    a given entry is in the file.
 
    Leave the following blank line there!!  Autoheader needs it.  */
-
+
 
 /* Define if the debug option is enabled. */
 #undef DEBUG
 
-/* Define if qsort is unstable. */
-#undef QSORT_UNSTABLE
+/* Define for building the dynamic loadable IM-modules */
+#undef USE_DYNAMIC
 
-/* Define if we have the mergesort function. */
-#undef HAVE_MERGESORT
+/* Define for dlopen() function */
+#undef HAVE_DLOPEN
 
-/* Define if we have the qsort function. */
-#undef HAVE_QSORT
+/* Define for gettext() function */
+#undef HAVE_GETTEXT
+
+/* Define for Linux system */
+#undef LINUX
+
+/* Define for FreeBSD system */
+#undef FREEBSD
+
+/* Define for OpenBSD system */
+#undef OPENBSD
+
+/* Define for Solaris system */
+#undef SOLARIS
+
+/* Define for HP-UX system */
+#undef HPUX
