@@ -67,6 +67,10 @@ gjc@world.std.com
    Release 3.5 5-MAY-97 fixes, plus win95 "compiler" to create exe files.
   */
 
+#ifdef HAVE_CONFIG_H
+#  include "config.h"
+#endif
+
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
@@ -77,12 +81,17 @@ gjc@world.std.com
 #include <time.h>
 #include <errno.h>
 
+#ifdef HPUX
+#  define _INCLUDE_POSIX_SOURCE
+#  include <sys/times.h>
+#endif
+
 #include "siod.h"
 #include "siodp.h"
 
 static void init_slib_version(void)
 {setvar(cintern("*slib-version*"),
-	cintern("$Id: slib.c,v 1.16 1997/12/05 14:44:44 gjc Exp $"),
+	cintern("$Id: slib.c,v 1.6 1999/10/27 12:42:52 thhsieh Exp $"),
 	NIL);}
 
 char * __stdcall siod_version(void)
@@ -169,13 +178,13 @@ char *siod_lib = SIOD_LIB_DEFAULT;
 void __stdcall process_cla(int argc,char **argv,int warnflag)
 {int k;
  char *ptr;
- static siod_lib_set = 0;
+ static int siod_lib_set = 0;
 #if !defined(vms)
  if (!siod_lib_set)
    {
 #ifdef WIN32
 	 if (argc > 0)
-	 {siod_lib = strdup(argv[0]);
+	 {siod_lib = (char *)strdup(argv[0]);
 	  siod_lib_set = 1;
 	  if ((ptr = strrchr(siod_lib,'\\')))
 		  ptr[1] = 0;}
@@ -1421,8 +1430,9 @@ LISP user_gc(LISP args)
  flag = no_interrupt(1);
  errjmp_ok = 0;
  old_status_flag = gc_status_flag;
- if NNULLP(args)
+ if NNULLP(args) {
    if NULLP(car(args)) gc_status_flag = 0; else gc_status_flag = 1;
+ }
  gc_mark_and_sweep();
  gc_status_flag = old_status_flag;
  errjmp_ok = 1;
@@ -1443,8 +1453,9 @@ long freelist_length(void)
  
 LISP gc_status(LISP args)
 {long n,m;
- if NNULLP(args) 
+ if NNULLP(args) {
    if NULLP(car(args)) gc_status_flag = 0; else gc_status_flag = 1;
+ }
  if (gc_kind_copying == 1)
    {if (gc_status_flag)
       put_st("garbage collection is on\n");
@@ -1542,7 +1553,9 @@ LISP leval(LISP x,LISP env)
       tmp = envlookup(x,env);
       if NNULLP(tmp) return(CAR(tmp));
       tmp = VCELL(x);
+/*
       if EQ(tmp,unbound_marker) err_ubv(x);
+*/
       return(tmp);
     case tc_cons:
       tmp = CAR(x);
@@ -2055,7 +2068,7 @@ int flush_ws(struct gen_readio *f,char *eoferr)
  commentp = 0;
  while(1)
    {c = GETC_FCN(f);
-    if (c == EOF) if (eoferr) err(eoferr,NIL); else return(c);
+    if (c == EOF) {if (eoferr) err(eoferr,NIL); else return(c);}
     if (commentp) {if (c == '\n') commentp = 0;}
     else if (c == ';') commentp = 1;
     else if (!isspace(c)) return(c);}}
@@ -2152,16 +2165,16 @@ LISP lreadtk(char *buffer,long j)
     if (flag) return(tmp);}
  if (*p == '-') p+=1;
  adigit = 0;
- while(isdigit(*p)) {p+=1; adigit=1;}
+ while(isdigit((int)(*p))) {p+=1; adigit=1;}
  if (*p=='.')
    {p += 1;
-    while(isdigit(*p)) {p+=1; adigit=1;}}
+    while(isdigit((int)(*p))) {p+=1; adigit=1;}}
  if (!adigit) goto a_symbol;
  if (*p=='e')
    {p+=1;
     if (*p=='-'||*p=='+') p+=1;
-    if (!isdigit(*p)) goto a_symbol; else p+=1;
-    while(isdigit(*p)) p+=1;}
+    if (!isdigit((int)(*p))) goto a_symbol; else p+=1;
+    while(isdigit((int)(*p))) p+=1;}
  if (*p) goto a_symbol;
  return(flocons(atof(buffer)));
  a_symbol:
@@ -2274,7 +2287,7 @@ LISP vload(char *ofname,long cflag,long rflag)
    ungetc(c,f);
  if ((start = strstr(buffer,key)))
    {for(end = &start[strlen(key)];
-	*end && isalnum(*end);
+	*end && isalnum((int)(*end));
 	++end);
     j = end - start;
     memmove(buffer,start,j);
