@@ -20,7 +20,7 @@ Array-hacking code moved to another source file.
 
 static void init_sliba_version(void)
 {setvar(cintern("*sliba-version*"),
-	cintern("$Id: sliba.c,v 1.10 1998/02/10 12:55:56 gjc Exp $"),
+	cintern("$Id: sliba.c,v 1.6 1999/10/27 12:42:52 thhsieh Exp $"),
 	NIL);}
 
 static LISP sym_plists = NIL;
@@ -1151,9 +1151,9 @@ LISP string2number(LISP x,LISP b)
     result = (double) value;}
  else if ((base >= 1) && (base <= 16))
    {for(result = 0.0;*str;++str)
-     if (isdigit(*str))
+     if (isdigit((int)(*str)))
        result = result * base + *str - '0';
-     else if (isxdigit(*str))
+     else if (isxdigit((int)(*str)))
        result = result * base + toupper(*str) - 'A' + 10;}
  else
    return(err("number base not handled",b));
@@ -1231,7 +1231,7 @@ static void init_base64_table(void)
  base64_decode_table = (char *) malloc(256);
  memset(base64_decode_table,-1,256);
  for(j=0;j<65;++j)
-   base64_decode_table[base64_encode_table[j]] = j;}
+   base64_decode_table[(int)base64_encode_table[j]] = (char)j;}
 
 #define BITMSK(N) ((1 << (N)) - 1)
 
@@ -1441,11 +1441,12 @@ LISP butlast(LISP l)
 {INTERRUPT_CHECK();
  STACK_CHECK(&l);
  if NULLP(l) err("list is empty",l);
- if CONSP(l)
+ if CONSP(l) {
    if NULLP(CDR(l))
      return(NIL);
    else
      return(cons(CAR(l),butlast(CDR(l))));
+ }
  return(err("not a list",l));}
 
 LISP nconc(LISP a,LISP b)
