@@ -1,6 +1,3 @@
-/*
- *  $Id: check_file_exist.c,v 1.1.1.2 1999/01/25 16:27:15 thhsieh Exp $
- */
 /*
     Copyright (C) 1999 by  XCIN TEAM
 
@@ -22,12 +19,18 @@
     xcin@linux.org.tw, or the maintainer Tung-Han Hsieh: thhsieh@linux.org.tw
 */      
 
+#ifdef HAVE_CONFIG_H
+#  include "config.h"
+#endif
 
 #include <unistd.h>
+#ifdef HPUX
+#  define _INCLUDE_POSIX_SOURCE
+#endif
 #include <sys/stat.h>
 #include "xcintool.h"
 
-int check_file_exist(char *path, enum ftype type)
+int check_file_exist(char *path, int type)
 {
     struct stat buf;
 
