@@ -1,6 +1,3 @@
-/*
- *  $Id: check_version.c,v 1.3 1999/02/18 08:09:44 thhsieh Exp $
- */
 /*
     Copyright (C) 1999 by  XCIN TEAM
 
@@ -22,6 +19,9 @@
     xcin@linux.org.tw, or the maintainer Tung-Han Hsieh: thhsieh@linux.org.tw
 */      
 
+#ifdef HAVE_CONFIG_H
+#  include "config.h"
+#endif
 
 #include <string.h>
 #include "xcintool.h"
