@@ -1,6 +1,3 @@
-/*
- *  $Id: copy_file.c,v 1.1.1.2 1999/01/25 16:27:17 thhsieh Exp $
- */
 /*
     Copyright (C) 1999 by  XCIN TEAM
 
@@ -22,6 +19,9 @@
     xcin@linux.org.tw, or the maintainer Tung-Han Hsieh: thhsieh@linux.org.tw
 */      
 
+#ifdef HAVE_CONFIG_H
+#  include "config.h"
+#endif
 
 #include <stdio.h>
 #include "xcintool.h"
