@@ -1,6 +1,3 @@
-/*
- *  $Id: open_file.c,v 1.2 1999/01/31 11:35:03 thhsieh Exp $
- */
 /*
     Copyright (C) 1999 by  XCIN TEAM
 
@@ -22,6 +19,9 @@
     xcin@linux.org.tw, or the maintainer Tung-Han Hsieh: thhsieh@linux.org.tw
 */      
 
+#ifdef HAVE_CONFIG_H
+#  include "config.h"
+#endif
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -34,6 +34,6 @@ open_file(char *fn, char *md, int exitcode)
 
     if ((f = fopen(fn, md)) == NULL && 
 	exitcode!=XCINMSG_EMPTY && exitcode!=XCINMSG_NORMAL)
-        perr(exitcode, _("cannot open file:  %s.\n"), fn);
+        perr(exitcode, N_("cannot open file:  %s.\n"), fn);
     return f;
 }
