@@ -26,15 +26,18 @@ typedef struct {
     char default_dir[256];
     char user_dir[256];
     char *lc_ctype;
+    char *encoding;
+    char *fname;
     char *fname_cin;
     char *fname_tab;
     char *sysfn;
     FILE *fr, *fw;
     int lineno;
-    char verbose;
+    byte_t verbose;
 } cintab_t;
 
 extern int cmd_arg(char *cmd, int cmdlen, ...);
 extern int read_hexwch(unsigned char *wch_str, char *arg);
+extern void load_systab(char *sysfn);
 
 #endif
